/* Set.c
 * Michael McTernan, mm7323@bris.ac.uk
 * 22/01/99
 * A two way linked list with a hash table
 *  structure allows fast insertions, deletions
 *  and lookups.
 * Only stores pointers, and avoids malloc once
 *  initialised by maintaining it's own free
 *  block list.
 * Datastructure for speed!
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "Set.h"

#define Error(a)    {fprintf(stderr,a);abort();}
#define Warning(a)  {fprintf(stderr,a);}

typedef struct element {
  void *Data,*Key;
  struct element *Chain,*Next,*Prev;
} element_t;

struct set {
  int Size;
  int KeyLength;
  int NumElements;
  element_t *FreeList,*List;
  element_t **Table;
  element_t *Iterator;
};


/*CalcHashCode*/
/* Simple hashing function */
inline static int CalcHashCode(set_t TheSet,void *Key) {
  int t,Code=0;

  for(t=0;t<TheSet->KeyLength;t++)
    Code+=((char *)Key)[t] * t;

  Code*=7;

  if(Code<0)
    Code=-Code;

  Code%=TheSet->Size;

  return 0;
}


/*SET_New*/
/* Create a new set.
 */
struct set *SET_New(int Size,int KeyLength) {
  struct set *NewSet;
  int t;

  if((NewSet=calloc(sizeof(struct set),1))==NULL)
    Error("calloc failed");

  NewSet->KeyLength=KeyLength;
  NewSet->Size=Size;

  if((NewSet->Table=calloc(sizeof(element_t),Size))==NULL)
    Error("calloc failed");

  /* Create the free list */
  for(t=0;t<Size/0.8;t++) {
    element_t *NewElement=calloc(sizeof(element_t),1);
   
    if(NewElement==NULL)
      Error("calloc failed");

    NewElement->Next=NewSet->FreeList;
    NewSet->FreeList=NewElement;

  } 

  return NewSet;
}


/*SET_Destroy*/
/* Destroy a set and free all memory 
 * (but not the data).
 */
void SET_Destroy(struct set *TheSet) {
  element_t *e;
  
  /* Free the free list */
  e=TheSet->FreeList;
  while(e!=NULL) {
    element_t *f;
    
    f=e->Next;
    free(e);
    e=f;
    
  }
    
  /* Free the list */
  e=TheSet->List;
  while(e!=NULL) {
    element_t *f;
    
    f=e->Next;
    free(e);
    e=f;
    
  }
  
  /* Free the table and the set */
  free(TheSet->Table);
  
  free(TheSet);

}


/*SET_Add*/
/* Take an element from the free list, fill in it's
 * fields and stitch it into the list and table.
 */
void SET_Add(struct set *TheSet,void *Data,void *Key) {
  element_t *NewElement;
  int HashCode;

  /* Increase the count */
  TheSet->NumElements++;

  /* Get a new element from the free list */
  if(TheSet->FreeList==NULL)
    Error("Out of free elements");

  NewElement=TheSet->FreeList;
  TheSet->FreeList=NewElement->Next;

  /* Setup the new element */
  NewElement->Data=Data;
  NewElement->Key=Key;

  /* Add to linked list */
  NewElement->Next=TheSet->List;
  if(NewElement->Next!=NULL)
    NewElement->Next->Prev=NewElement;
  NewElement->Prev=NULL;
  TheSet->List=NewElement;

  /* Add to the hash table */
  HashCode=CalcHashCode(TheSet,Key);
  NewElement->Chain=(TheSet->Table)[HashCode];
  (TheSet->Table)[HashCode]=NewElement;

  return;
}


/*SET_Remove*/
/* Remove an entity from the database.
 */
void SET_Remove(set_t TheSet,void *Key) {
  element_t *Element,*PrevElement;

  /* Find the element from the hash table */
  PrevElement=NULL; 
  Element=(TheSet->Table)[CalcHashCode(TheSet,Key)];

  if(Element==NULL) {
    Warning("SET_Remove: Remove element not found in set");
    return;
  }

  while(memcmp(Element->Key,Key,TheSet->KeyLength)!=0) {
    PrevElement=Element;
    Element=Element->Chain;

    if(Element==NULL) {
      Warning("SET_Remove: Remove element not found in set");
      return;
    }
  }

  /* Remove from the hash table */
  if(PrevElement==NULL)
    (TheSet->Table)[CalcHashCode(TheSet,Key)]=Element->Chain;
  else 
    PrevElement->Chain=Element->Chain;

  /* Check that we are not about to break the iterator */
  if(Element==TheSet->Iterator && Element!=NULL) {
    TheSet->Iterator=PrevElement;
  }


  /* Remove the element from the list */
  if(Element==TheSet->List) {
    TheSet->List=Element->Next;
    if(Element->Next!=NULL)
      Element->Next->Prev=NULL;
  } else {
    if(Element->Next!=NULL)
      Element->Next->Prev=Element->Prev;
    Element->Prev->Next=Element->Next;
  }

  /* Now add the element to the free list */
  Element->Next=TheSet->FreeList;
  TheSet->FreeList=Element;

  /* Decrease the count */
  TheSet->NumElements--;

}


/*SET_Get*/
/* Get an item from the set.
 */
void *SET_Get(set_t TheSet,void *Key) {
  element_t *Element;

  /* Find the element from the hash table */
  Element=(TheSet->Table)[CalcHashCode(TheSet,Key)];

  if(Element==NULL)
    return NULL;
  
  while(memcmp(Element->Key,Key,TheSet->KeyLength)!=0) {
    Element=Element->Chain;

    if(Element==NULL)
      return NULL;
  };

  return Element->Data;

}


/*SET_Size*/
/* Return the number of elements in the set
 */
int SET_Size(set_t TheSet) {
  return TheSet->NumElements;
}


/*SET_Iterate*/
/* Return an item from the list, or NULL
 * if we have gone through them all.
 */
void *SET_Iterate(set_t TheSet) {

  if(TheSet->Iterator==NULL) {
    TheSet->Iterator=TheSet->List;
  } else {
    TheSet->Iterator=TheSet->Iterator->Next;
  }

  if(TheSet->Iterator!=NULL)
    return TheSet->Iterator->Data;

  return NULL;
  
}

/*SET_ResetIterator*/
/* Set the iterator to the start
 * of the list
 */
void SET_ResetIterator(set_t TheSet) {
  TheSet->Iterator=NULL;
}
